#ifndef X_COMM_HPP
#define X_COMM_HPP

#include <conio.h>
#include <malloc.h>
#include "x_types.h"

#define IN(x)		inp(base + x)		// read a UART register
#define OUT(x,y)	outp(base + x, y)	// write a UART register
#define RBR		0		// receive buffer register
#define THR		0		// transmit holding register
#define DLL		0		// divisor latch LSB
#define DLM		1		// divisor latch MSB
#define IER		1		// interrupt enable register
#define IIR		2		// interrupt id register
#define FCR		2		// FIFO control register
#define AFR		2		// alternate function register
#define LCR		3		// line control register
#define MCR		4		// modem control register
#define LSR		5		// line status register
#define MSR		6		// modem status register
#define SCR		7		// scratch register
					// interrupt enable register
#define IER_RBF		0x01		//  receive buffer full
#define IER_TBE		0x02		//  transmit buffer empty
#define IER_LSI		0x04		//  line status interrupt
#define IER_MSI		0x08		//  modem status interrupt
#define IER_ALL		0x0f		//  enable all interrupts

					// interrupt id register
#define IIR_PEND	0x01		//  interrupt pending = 0
#define IIR_II		0x06		//  interrupt id bits
					//   000 = modem status change
					//   001 = trans holding empty
					//   010 = receive buffer full
					//   110 = receive fifo full
					//   011 = line status change
#define IIR_MSI		0x00		//  modem status interrupt
#define IIR_TBE		0x02		//  transmit buffer empty
#define IIR_RBF		0x04		//  receive buffer full
#define IIR_LSI		0x06		//  line status interrupt
#define IIR_RFF		0x0c		//  receive fifo threshold

					// fifo control register
#define FCR_FIFO	0x01		//  fifo enable
#define FCR_RCVR	0x02		//  receiver fifo reset
#define FCR_XMIT	0x04		//  transmit fifo reset
#define FCR_DMA		0x08		//  DMA mode select
#define FCR_TRIGGER	0xc0		//  receiver trigger select
					//   00 = 1 byte
					//   01 = 4 bytes
					//   10 = 8 bytes
					//   11 = 14 bytes

					// line control register
#define LCR_WLEN	0x03		//  word length
					//   10 = 7 bits
					//   11 = 8 bits
#define LCR_STOP	0x04		//  stop bits
					//   0 = 1 stop bit
					//   1 = 2 stop bits
#define LCR_PARITY	0x08		//  parity enable
					//   0 = no parity
					//   1 = send/check parity
#define LCR_EVEN	0x10		//  even/odd parity
					//   0 = odd/mark parity
					//   1 = even/space parity
#define LCR_STICK	0x20		//  stick parity select
#define LCR_BREAK	0x40		//  break, set to xmit break
#define LCR_DLAB	0x80		//  divisor latch access bit

					// modem control register
#define MCR_DTR		0x01		//  DTR control
#define MCR_RTS		0x02		//  RTS control
#define MCR_OUT2	0x08		//  OUT2 control
#define MCR_DO		0x0b		//  DTR, RTS & OUT2 enabled

					// line status register
#define LSR_DR		0x01		//  data ready
#define LSR_ORUN	0x02		//  overrun error
#define LSR_PRTY	0x04		//  parity error
#define LSR_FRM		0x08		//  framing error
#define LSR_BRK		0x10		//  break interrupt
#define LSR_THRE	0x20		//  transmit holding reg empty
#define LSR_TSRE	0x40		//  transmit shift reg empty
#define LSR_ERROR	0x1e		//  error conditions

					// modem status register
#define MSR_DCTS	0x01		//  delta clear to send
#define MSR_DDSR	0x02		//  delta data set ready
#define MSR_TERI	0x04		//  trailing edge ring indicator
#define MSR_DCD		0x08		//  delta carrier detect
#define MSR_CTS		0x10		//  clear to send
#define MSR_DSR		0x20		//  data set ready (modem ready)
#define MSR_RI		0x40		//  ring indicated
#define MSR_CD		0x80		//  carrier detected

#define I8259		0x20		// control register address
#define I8259M		0x21		// mask register
#define EOI		0x20		// end of interrupt command

class Comm {
public:
  Comm(int b = 0x2f8,			// base address
       int i = 3,			// interrupt number
       long s = 38400,			// line speed
       int l = 3,			// line control setting
       int fc = 1,			// flow control flag
       int safety = 0,			// fifo safety (0-3)
       word si = 4096,			// input queue size
       word so = 4096);			// output queue size
  ~Comm();				// destructor
  word read(char *c);			// read a character (no status)
  word read(char *c,			// read a character from queue
	    char *m,			// get modem status register
	    char *l);			// get line status register
  void write(int c);			// write a character
  void write(char *s);			// write a string
  void write(char *s, int l);		// write a substring

  void command(char *c, char *r = NULL,	// send command to modem
	       void (*idle)(int) = NULL);
  word getIcount();			// get depth of input queue
  word getOcount();			// get depth of output queue
  word iFree();				// free space in input queue
  word oFree();				// free space in output queue
  long getBPS();			// rtn current speed in bps
  byte getMSR();			// rtn modem status register
  byte getLCR();			// rtn line control register
  int UART_16550();			// rtn TRUE if 16550 detected
  int carrier();			// rtn TRUE if carrier detected
  int getMSRchanged();			// rtn TRUE if msr changed
  int getIflow();			// rtn input flow ctrl status
  int iEmpty();				// rtn TRUE if input queue empty
  int oEmpty();				// rtn TRUE if output queue empty
  void iClear();			// clear input queue
  void oClear();			// clear output queue
  word set8N();				// set 8 data bits, no parity
  void setBPS(long l, int safety);	// set port speed
  void setLCR(int l);			// set up LCR
  void lowerDTR(word t = 500,		// lower DTR temporarily
	   void (*idle)(int) = NULL);
  void setRTS(int f);			// RTS signal control
  void ISR();				// COMM ISR
protected:
  word base,				// base address
       irq,				// interrupt number
       divisor,				// baud rate divisor
       line,				// initial line control
       iSize,				// input buffer size
       oSize,				// output buffer size
       iStart,				// flow ctl restart limit
       iStop,				// flow ctl upper stop limit
       iCount,				// chars in input queue
       oCount,				// chars in output queue
       deQ();				// deque an output queue char
  char *iBuf,				// input buffer
       *iGet,				// ..next get location
       *iPut,				// ..next put location
       *iLimit,				// ..last location
       *iLast,				// ..last iPut location
       iOverflow,				// ..input overflow flag
       *oBuf,				// output buffer
       *oGet,				// ..next get location
       *oPut,				// ..next put location
       *oLimit,				// ..last location
       MSRchanged,			// MSR changed flag
       intMSR,				// last interrupt MSR
       intLSR,				// last interrupt LSR
       fifo,				// 16550 flag
       flow,				// flow control enable flag
       oFlow,				// output flow controlled flag
       iFlow,				// input flow controlled flag
       emptyTrans;			// empty transmitter flag
  void __interrupt (*oldComm)();	// old comm ISR pointer
  void installISR();			// install ISR
  void uninstallISR();			// de-install ISR
  void detectBPS(int safety);		// autodetect DTE speed
  void setLimits();			// set up flow ctl limits
  void checkFIFO();			// set up fifo flags
  void enQ(int c,			// queue char to input queue
	   int m,			// save MSR
	   int l);			// save LSR
};

extern Comm *comm;			// current Comm instance
extern void __interrupt commISR();	// comm ISR

#endif
